/*
  File name        : cpumon.c
  Date of creation : 4/1/2009
  Version          : 2.6
  Copyright        : Soft'n'design
  Author           : Laurent Mazet <mazet@softndesign.org>

  Description      : This file contains cpu monitoring

  History          :
  - initial version
*/

/* check OS */
#ifdef __FREEBSD__
#elif __LINUX__
#elif __NETBSD__
#elif __SYLLABLE__
#elif __DEPEND__
#else
#error "unsupported operating system"
#endif

#define _GNU_SOURCE

#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#ifdef __FREEBSD__
#include <pthread_np.h>
#endif
#include <sched.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef  __FREEBSD__
#include <sys/cpuset.h>
#elif __NETBSD__
#include <sys/sched.h>
#endif
#ifdef __SYLLABLE__
typedef cpu_set_t cpu_set;
#endif
#include <sys/param.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "rdtsc.h"
#include "verbose.h"

#define VERSION "2.6"

/**
   Default values
*/
int cpu_cal = -1;
int cpu_mon = 0;
#if defined (__FREEBSD__) || defined (__NETBSD__)
#define CALIBRATION_FILE "/var/cpumon/calibration.txt"
#elif __LINUX__
#define CALIBRATION_FILE "/var/lib/cpumon/calibration.txt"
#elif __SYLLABLE__
#define CALIBRATION_FILE "/var/lib/cpumon.txt"
#endif
char *calibration_file = CALIBRATION_FILE;
int nb_calibration_loops = 1;
int stop_after_nb_seconds = -1;
int timer = 5;
DECLARE_VERBOSE_LEVEL (cpu, 1);

/**
   Global variables
*/
double max_clock = -1;
double current_clock = -1;
int max_loops = -1;
int nb_cpu = -1;
pid_t *pids = NULL;
int *pipes = NULL;
char *progname = NULL;

/**
   Display help message

   @param ret return code
*/
void usage (int ret)
{
    printf ("usage: %s [-c int] [-f str] [-h] [-k int] [-m int] [-s int] [-t int] [-v int] [-V]\n", progname);
    printf (" -c int: force calibration (%d), need root\n", nb_calibration_loops);
    printf (" -f str: calibration file (%s)\n", calibration_file);
    printf (" -h    : help message\n");
    printf (" -k int: cpu id for calibration, -1 means last core (%d)\n", cpu_cal);
    printf (" -m int: cpu id for monitoring, -1 means last core (%d)\n", cpu_mon);
    printf (" -s int: number of second before stopping, -1 means infinity (%d)\n", stop_after_nb_seconds);
    printf (" -t int: timer period (%d)\n", timer);
    printf (" -v int: verbose level (%d)\n", verbose_cpu);
    printf (" -V    : show version\n");

    exit (ret);
}

/**
   Count number of CPU

   @return number of CPU
*/
int get_cpu_number (void)
{
    static int ncpu = 0;

    if (ncpu == 0) {
        int i, _ncpu = 0;
#ifdef __NETBSD__
        cpuset_t *cpu_set;
#elif __FREEBSD__
        cpuset_t cpu_mask;
#elif __SYLLABLE__
        cpu_set cpu_mask;
#elif __LINUX__
        cpu_set_t cpu_mask;
#endif

        /* test first core */
#ifdef __NETBSD__
        cpu_set = cpuset_create ();
        if (cpu_set == NULL) {
            VERBOSE (cpu, ERROR, printf ("cpuset_create: %s\n", strerror (errno)));
            exit (1);
        }
        cpuset_set (0, cpu_set);
        if (pthread_setaffinity_np (pthread_self (), cpuset_size (cpu_set), cpu_set) != 0) {
            VERBOSE (cpu, ERROR, printf ("pthread_setaffinity: %s\n", strerror (errno)));
            exit (1);
        }
#elif __FREEBSD__
        CPU_ZERO (&cpu_mask);
        CPU_SET (0, &cpu_mask);
        if (pthread_setaffinity_np (pthread_self (), sizeof (cpuset_t), &cpu_mask) != 0) {
            VERBOSE (cpu, ERROR, printf ("pthread_setaffinity: %s\n", strerror (errno)));
            exit (1);
        }
#elif __SYLLABLE__
        __CPU_ZERO (&cpu_mask);
        __CPU_SET (0, &cpu_mask);
        if (sched_setaffinity (getpid (), sizeof (cpu_set), &cpu_mask) != 0) {
            /* Function not implemented */
            VERBOSE (cpu, ERROR, printf ("sched_setaffinity: %s\n", strerror (errno)));
            //exit (1);
        }
#elif __LINUX__
        CPU_ZERO (&cpu_mask);
        CPU_SET (0, &cpu_mask);
        if (pthread_setaffinity_np (pthread_self (), sizeof (cpu_set_t), &cpu_mask) != 0) {
            VERBOSE (cpu, ERROR, printf ("pthread_setaffinity: %s\n", strerror (errno)));
            exit (1);
        }
#endif
        _ncpu++;

        /* count number of cpu */

#if defined (__FREEBSD__) || defined (__LINUX__) || defined (__SYLLABLE__)
       int max_number_cpu = CPU_SETSIZE;
#elif __NETBSD__
       int max_number_cpu = cpuset_size (cpu_set) * 8;
#endif

        for (i = 1; i < max_number_cpu; i++) {

#ifdef __NETBSD__
            cpuset_zero (cpu_set);
            cpuset_set (i, cpu_set);
            if ((pthread_setaffinity_np (pthread_self (), cpuset_size (cpu_set), cpu_set) != 0) ||
                (pthread_getaffinity_np (pthread_self (), cpuset_size (cpu_set),
 cpu_set) != 0) ||
                (!cpuset_isset (i, cpu_set)))
                break;
#elif __FREEBSD__
            CPU_ZERO (&cpu_mask);
            CPU_SET (i, &cpu_mask);
            if ((pthread_setaffinity_np (pthread_self (), sizeof (cpuset_t), &cpu_mask) != 0) ||
                (pthread_getaffinity_np (pthread_self (), sizeof (cpuset_t), &cpu_mask) != 0) ||
                (!CPU_ISSET (i, &cpu_mask)))
                break;
#elif __SYLLABLE__
            __CPU_ZERO (&cpu_mask);
            __CPU_SET (i, &cpu_mask);
            if ((sched_setaffinity (getpid (), sizeof (cpu_set), &cpu_mask) != 0) ||
                (sched_getaffinity (getpid (), sizeof (cpu_set), &cpu_mask) != 0) ||
                (!__CPU_ISSET (i, &cpu_mask)))
                break;
#elif __LINUX__
            CPU_ZERO (&cpu_mask);
            CPU_SET (i, &cpu_mask);
            if ((pthread_setaffinity_np (pthread_self (), sizeof (cpu_set_t), &cpu_mask) != 0) ||
                (pthread_getaffinity_np (pthread_self (), sizeof (cpu_set_t), &cpu_mask) != 0) ||
                (!CPU_ISSET (i, &cpu_mask)))
                break;
#endif

            _ncpu++;
        }

        ncpu = _ncpu;
        VERBOSE (cpu, DEBUG, printf ("got %d cores\n", ncpu));

        /* restore affinity for all cores */
#ifdef __NETBSD__
        cpuset_zero (cpu_set);
        for (i = 0; i < ncpu; i++)
            cpuset_set (i, cpu_set);
        if (pthread_setaffinity_np (pthread_self (), cpuset_size (cpu_set), cpu_set) != 0) {
            VERBOSE (cpu, ERROR, printf ("pthread_setaffinity: %s\n", strerror (errno)));
            //exit (1);
        }
#elif __FREEBSD__
        CPU_ZERO (&cpu_mask);
        for (i = 0; i < ncpu; i++)
            CPU_SET (i, &cpu_mask);
        if (pthread_setaffinity_np (pthread_self (), sizeof (cpuset_t), &cpu_mask) != 0) {
            VERBOSE (cpu, ERROR, printf ("pthread_setaffinity: %s\n", strerror (errno)));
            exit (1);
        }
#elif __SYLLABLE__
        __CPU_ZERO (&cpu_mask);
        for (i = 0; i < ncpu; i++)
            __CPU_SET (i, &cpu_mask);
        if (sched_setaffinity (getpid (), sizeof (cpu_set), &cpu_mask) != 0) {
            /* Function not implemented */
            VERBOSE (cpu, ERROR, printf ("sched_setaffinity: %s\n", strerror (errno)));
            //exit (1);
        }
#elif __LINUX__
        CPU_ZERO (&cpu_mask);
        for (i = 0; i < ncpu; i++)
            CPU_SET (i, &cpu_mask);
        if (pthread_setaffinity_np (pthread_self (), sizeof (cpu_set_t), &cpu_mask) != 0) {
            VERBOSE (cpu, ERROR, printf ("pthread_setaffinity: %s\n", strerror (errno)));
            exit (1);
        }
#endif

#ifdef __NETBSD__
        cpuset_destroy (cpu_set);
#endif
    }

    return ncpu;
}

/**
   Set new name to distinguish forked process

   @param argc number of arguments
   @param argv tabular of argument strings
   @param name new processus name
*/
void set_name (int argc, char *argv[], char *name)
{
    int n = (int)(argv[argc - 1] - argv[0]) + strlen (argv[argc - 1]) + 1;
    memset (argv[0], 0, n);
    strncpy (argv[0], name, n);
}

/**
   Handler for sigchild

   @param signal
*/
void  sigchld_handler (int signal)
{
    if (signal == SIGCHLD) {
        int status = 0;
        VERBOSE (cpu, WARNING, printf ("receive sigchild\n"));
        wait (&status);
    }
}

/**
   Handler for sigpipe

   @param signal
*/
void  sigpipe_handler (int signal)
{
    if (signal == SIGPIPE)
        VERBOSE (cpu, ERROR, printf ("receive sigpipe\n"));
}

/**
   Alarm trigger
*/
int alarm_trigger = 0;

/**
   Handler for sigalarm

   @param signal
*/
void sigalrm_handler (int sig)
{
    if (sig == SIGALRM)
        alarm_trigger = 1;
}

/**
   Kill all children at exit
*/
void kill_all_children (void)
{
    int i;
    for (i = 0; i < nb_cpu; i++)
        kill (pids[i], SIGKILL);
}

/**
   Compute number of loops per second

   @param cpu identifier

   @return number of loops during 1s
*/
int loop_per_second (int cpu, int *cal)
{

    /* get time */
    struct timeval tv1;
    gettimeofday (&tv1, NULL);
    int64_t ts1 = rdtsc ();

    /* set alarm */
    alarm_trigger = 0;
    alarm (1);

    /* loop counter */
    int cpt = -1;
    int64_t min_ts = INT64_MAX;
    while (!alarm_trigger) {
        int k, sum = 0;
        int64_t ts = rdtsc ();
        for (k = 0;k < 10000; k++) sum += k;
        ts = rdtsc () - ts;
        if (min_ts > ts)
            min_ts = ts;
        cpt++;
    }

    /* get time */
    int64_t ts2 = rdtsc ();
    struct timeval tv2;
    gettimeofday (&tv2, NULL);

    /* cpu clock estimate */
    double delta = ((double)tv2.tv_sec + tv2.tv_usec * 1e-6) -
        ((double)tv1.tv_sec + tv1.tv_usec * 1e-6);
    double current_clock = (ts2 - ts1) / delta;
    if (cal)
        *cal = (ts2 - ts1) / min_ts;
    VERBOSE (cpu, DEBUG, printf ("cpu%d: %.3fGHz %d, loop in %.0f ms\n",
                                 cpu, current_clock * 1e-9, cpt, delta * 1e3));

    return cpt;
}

/**
   Assign a job to a cpu

   @param cpu identifier
*/
void set_cpu (int cpu)
{

#ifdef __NETBSD__
    cpuset_t *cpu_set;
    cpu_set = cpuset_create ();
    if (cpu_set == NULL) {
        VERBOSE (cpu, ERROR, printf ("cpuset_create: %s\n", strerror (errno)));
        exit (1);
    }
    cpuset_set (cpu, cpu_set);
    if (pthread_setaffinity_np (pthread_self (), cpuset_size (cpu_set), cpu_set) != 0) {
        VERBOSE (cpu, ERROR, printf ("pthread_setaffinity: %s\n", strerror (errno)));
        exit (1);
    }
#elif __FREEBSD__
    cpuset_t cpu_mask;
    CPU_ZERO (&cpu_mask);
    CPU_SET (cpu, &cpu_mask);
    if (pthread_setaffinity_np (pthread_self (), sizeof (cpuset_t), &cpu_mask) != 0) {
        VERBOSE (cpu, ERROR, printf ("pthread_setaffinity: %s\n", strerror (errno)));
        exit (1);
    }
#elif __SYLLABLE__
    cpu_set cpu_mask;
    __CPU_ZERO (&cpu_mask);
    __CPU_SET (cpu, &cpu_mask);
    if (sched_setaffinity (getpid (), sizeof (cpu_set), &cpu_mask) != 0) {
        /* Function not implemented */
        VERBOSE (cpu, ERROR, printf ("sched_setaffinity: %s\n", strerror (errno)));
        //exit (1);
    }
#elif __LINUX__
    cpu_set_t cpu_mask;
    CPU_ZERO (&cpu_mask);
    CPU_SET (cpu, &cpu_mask);
    if (pthread_setaffinity_np (pthread_self (), sizeof (cpu_set_t), &cpu_mask) != 0) {
        VERBOSE (cpu, ERROR, printf ("pthread_setaffinity: %s\n", strerror (errno)));
        exit (1);
    }
#endif

#ifdef __NETBSD__
    cpuset_destroy (cpu_set);
#endif

}

/**
   Set scheduler to realtime at high priority

   @param step diff to highest priority

   @return 0 on error otherwise 1
*/
int set_high_priority (int step)
{
    struct sched_param sp;
    memset (&sp, 0, sizeof (sp));
    sp.sched_priority = sched_get_priority_max (SCHED_FIFO) - ((step > 0) ? step : 1);
    VERBOSE (cpu, DEBUG, printf ("SCHED_FIFO (%d, %d) SCHED_OTHER (%d, %d)\n",
                sched_get_priority_min (SCHED_FIFO), sched_get_priority_max (SCHED_FIFO),
                sched_get_priority_min (SCHED_OTHER), sched_get_priority_max (SCHED_OTHER)));
    VERBOSE (cpu,DEBUG, printf ("change scheduler from %d to SCHED_FIFO (%d) priority %d\n", sched_getscheduler (getpid ()), SCHED_FIFO, sp.sched_priority));
    if (sched_setscheduler (0, SCHED_FIFO, &sp) < 0) {
        VERBOSE (cpu, WARNING, printf ("sched_setscheduler (SCHED_FIFO, {%d}): %s\n", sp.sched_priority, strerror (errno)));
        return 0;
    }
    VERBOSE (cpu, DEBUG, printf ("sched_getscheduler = %d\n", sched_getscheduler (getpid ())));
    VERBOSE (cpu, DEBUG, printf ("running high realtime priority\n"));

    return 1;
}

/**
   Set low priority

   @param step diff to lowest priority
*/
void set_low_priority (int step)
{
    struct sched_param sp;
    memset (&sp, 0, sizeof (sp));
#if defined (__FREEBSD__) || defined (__LINUX__) || defined (__SYLLABLE__)
    sp.sched_priority = sched_get_priority_max (SCHED_OTHER);
#elif __NETBSD__
    sp.sched_priority = PRI_NONE;
#endif
    VERBOSE (cpu, DEBUG, printf ("SCHED_FIFO (%d, %d) SCHED_OTHER (%d, %d)\n",
                sched_get_priority_min (SCHED_FIFO), sched_get_priority_max (SCHED_FIFO),
                sched_get_priority_min (SCHED_OTHER), sched_get_priority_max (SCHED_OTHER)));
    VERBOSE (cpu,DEBUG, printf ("change scheduler from %d to SCHED_OTHER (%d) priority %d\n", sched_getscheduler (getpid ()), SCHED_OTHER, sp.sched_priority));
    if (sched_setscheduler (getpid (), SCHED_OTHER, &sp) < 0) {
        VERBOSE (cpu, ERROR, printf ("sched_setscheduler (SCHED_OTHER, {%d}): %s\n", sp.sched_priority, strerror (errno)));
        sched_getparam (getpid (), &sp);
        exit (0);
    }
    VERBOSE (cpu, DEBUG, printf ("sched_getscheduler = %d\n", sched_getscheduler (getpid ())));
    if (nice (20 - step) < 0) {
        VERBOSE (cpu, ERROR, printf ("nice (%d): %s\n", 20 - step, strerror (errno)));
        exit (1);
    }
    VERBOSE (cpu, DEBUG, printf ("running low priority\n"));
}

/**
   Calibration

   @param cpu identifier

   @return number of loops
*/
int calibrate (int cpu)
{

    /* check arguments */
    if ((cpu < 0) || ( cpu >= nb_cpu)) {
        VERBOSE (cpu, ERROR, printf ("incorrect cpu id (%d)\n", cpu));
        exit (1);
    }
    if (nb_calibration_loops <= 0) {
        VERBOSE (cpu, ERROR, printf ("incorrect number of calibration loops (%d)\n", nb_calibration_loops));
        exit (1);
    }

    /* assign to a cpu */
    set_cpu (cpu);

    /* set scheduler to realtime at highest prioriry */
    if (!set_high_priority (1)) {
        VERBOSE (cpu, WARNING, printf ("must be root to set scheduler to realtime at highest prioriry\n"));
    }

    /* alarm signal handler */
    struct sigaction act;
    memset (&act, 0, sizeof (act));
    act.sa_handler = sigalrm_handler;
    if (sigaction (SIGALRM, &act, NULL))
        VERBOSE (cpu, WARNING, printf ("sigaction (SIGALRM)\n"));

    /* calibration loop */
    int i, j;
    int *nbloops = (int *) calloc (nb_calibration_loops, sizeof (int));
    for (i = 0; i < nb_calibration_loops; i++) {
        int m;
        m = loop_per_second (cpu, &m);
        //if (!m) continue;
        for (j = 0; j < i; j++) {
            if (m > nbloops[j])
                break;
        }
        if (j != i)
            memmove (nbloops + j + 1, nbloops + j, (i - j) * sizeof (int));
        nbloops[j] = m;
    }
    VERBOSE (cpu, DEBUG, int i; printf ("nbloops:"); for (i = 0; i < nb_calibration_loops; i++) printf (" %d", nbloops[i]); printf ("\n"));

    /* get median value */
    int median = nbloops[nb_calibration_loops / 2];
    free (nbloops);
    VERBOSE (cpu, DEBUG, printf ("calibrate %d loops\n", median));

    return median;
}

/**
   Read calibration file

   @param fname file name

   @return maximum number of loops (-1 on error)
*/
int read_calibration_file (char *fname)
{
    int nbloops = -1;

    FILE *fd = fopen (fname, "r");
    if (fd) {
        char buffer[1024] = { 0 };
        char *ptr;
        fgets (buffer, sizeof (buffer), fd);
        nbloops = strtol (buffer, &ptr, 10);
        current_clock = atof (ptr);
        fclose (fd);
        VERBOSE (cpu, DEBUG, printf ("read calibration file (%s) %d / %.3fGHz\n", fname, nbloops, current_clock));
    } else {
        VERBOSE (cpu, WARNING, printf ("can't read calibration file (%s)\n", fname));
    }

    return nbloops;
}

/**
   Write calibration file

   @param fname file name
*/
void write_calibration_file (char *fname)
{

    /* check if directory exists otherwise create it */
    char *calibration_dir = dirname (strdup (fname));
    DIR *dd = opendir (calibration_dir);
    if (!dd) {
        if (mkdir (calibration_dir, 0755)) {
            VERBOSE (cpu, WARNING, printf ("can't create directory (%s)\n", calibration_dir));
        }
    } else {
        closedir (dd);
    }

    /* write calibration file */
    FILE *fd = fopen (fname, "w");
    if (fd) {
        fprintf (fd, "%d %g", max_loops, max_clock);
        fclose (fd);
        VERBOSE (cpu, DEBUG, printf ("write calibration file (%s)\n", fname));
    } else {
        VERBOSE (cpu, WARNING, printf ("can't write calibration file (%s)\n", fname));
    }

}

/**
   Handler for sigint

   @param signal
*/
void sigint_handler (int sig)
{
    if (sig == SIGINT) {
        VERBOSE (cpu, WARNING, printf ("ending cpumon processes\n"));
        if (max_loops > 0)
            write_calibration_file (calibration_file);
        exit (0);
    }
}

/**
   Measurement process

   @param cpu identifier
*/
void measurement_process (int cpu)
{

    /* check arguments */
    if ((cpu < 0) || ( cpu >= nb_cpu)) {
        VERBOSE (cpu, ERROR, printf ("incorrect cpu id (%d)\n", cpu));
        exit (1);
    }

    /* assign to a cpu */
    set_cpu (cpu);

    /* set scheduler to other at low prioriry */
    set_low_priority (0);

    /* alarm signal handler */
    struct sigaction act;
    memset (&act, 0, sizeof (act));
    act.sa_handler = sigalrm_handler;
    if (sigaction (SIGALRM, &act, NULL))
        VERBOSE (cpu, WARNING, printf ("sigaction (SIGALRM)\n"));

    /* measurement loop */
    while (1) {
        int mes = loop_per_second (cpu, NULL);
        VERBOSE (cpu, DEBUG, printf ("mes%d %d\n", cpu, mes));

        /* ensure that measure is always less than max_loops */
        if (max_loops < mes) {
            max_loops = mes;
            mes = -mes;
        }

        /* write measure through pipe to monitor */
        if (write (pipes[cpu], &mes, sizeof (mes)) != sizeof (mes)) {
            VERBOSE (cpu, ERROR, printf ("can't write in pipe for cpu-%02d\n", cpu));
            exit (1);
        }
        VERBOSE (cpu, DEBUG, printf ("write to %d\n", pipes[cpu]));
    }
}

/**
   Monitoring process

   @param cpu identifier
*/
void monitoring_process (int cpu)
{
    int i;

    /* check arguments */
    if ((cpu < 0) || ( cpu >= nb_cpu)) {
        VERBOSE (cpu, ERROR, printf ("incorrect cpu id (%d)\n", cpu));
        exit (1);
    }

    /* assign to a cpu */
    set_cpu (cpu);

    /* set scheduler to other at low prioriry */
    set_low_priority (1);

    /* alarm signal handler */
    struct sigaction act;
    memset (&act, 0, sizeof (act));
    act.sa_handler = sigalrm_handler;
    if (sigaction (SIGALRM, &act, NULL))
        VERBOSE (cpu, WARNING, printf ("sigaction (SIGALRM)\n"));

    /* results */
    float *measurements = (float *) calloc (nb_cpu, sizeof (float));
    int *counts = (int *) calloc (nb_cpu, sizeof (int));

    VERBOSE (cpu, DEBUG, printf ("pipes:"); for (i = 0; i < nb_cpu; i++) printf (" %d", pipes[i]); printf ("\n"));

    /* set alarm */
    int nb_seconds = 0;
    alarm_trigger = 0;
    alarm (timer);
    int64_t ts1 = rdtsc ();
    struct timeval tv1;
    gettimeofday (&tv1, NULL);

    while (1) {

        /* fids for select */
        fd_set rfds;
        FD_ZERO (&rfds);
        int fd_max = -1;
        for (i = 0; i < nb_cpu; i++) {
            FD_SET(pipes[i], &rfds);
            if (fd_max < pipes[i])
                fd_max = pipes[i];
        }

        /* timer */
        struct timeval tv;
        tv.tv_sec = 2 * timer;
        tv.tv_usec = 0;
        static int select_error = 0;

        /* wait for inputs */
        switch (select (fd_max + 1, &rfds, NULL, NULL, &tv)) {
        case -1:
            if (select_error++)
                VERBOSE (cpu, WARNING, printf ("select: %s\n", strerror (errno)));
            errno = 0;
            break;

        case 0:
            VERBOSE (cpu, WARNING, printf ("no data within %d seconds\n", 2 * timer));
            break;

        default:
            for (i = 0; i < nb_cpu; i++) {
                if (FD_ISSET (pipes[i], &rfds)) {
                    VERBOSE (cpu, DEBUG, printf ("read from %d\n", pipes[i]));

                    /* read measurement frm pipe */
                    int mes;
                    if (read (pipes[i], &mes, sizeof (mes)) != sizeof (mes)) {
                        VERBOSE (cpu, ERROR, printf ("can't read from %d\n", pipes[i]));
                        exit (1);
                    }

                    /* update calibration */
                    if (mes < 0) {
                        mes = -mes;
                        if (max_loops < mes) {
                            max_loops = mes;
                            VERBOSE (cpu, WARNING, printf ("update calibration to %d from cpu %d\n", max_loops, i));
                        }
                    }

                    /* compute remaing load */
                    float weight = 1;
                    if ((current_clock > 0) && (current_clock < max_clock * 0.99))
                        weight = current_clock / max_clock;
                    if (weight != 1)
                        VERBOSE (cpu, DEBUG, printf ("weight: %g\n", weight));
                    measurements[i] += (max_loops * weight - mes) / (max_loops * weight);
                    counts[i]++;
                }
            }
            break;
        }

        /* display monitoring information */
        if (alarm_trigger) {

            /* cpu clock estimate */
            int64_t ts2 = rdtsc ();
            struct timeval tv2;
            gettimeofday (&tv2, NULL);

            /* cpu clock estimate */
            double delta = ((double)tv2.tv_sec + tv2.tv_usec * 1e-6) -
                ((double)tv1.tv_sec + tv1.tv_usec * 1e-6);
            current_clock = (ts2 - ts1) / delta;
            if ((current_clock > 0) && ((max_clock > current_clock ? max_clock - current_clock : current_clock - max_clock) > 0.01f * current_clock))
                VERBOSE (cpu, WARNING, printf ("cpu speed is changing from %.3f to %.3f\n", max_clock * 1e-9, current_clock * 1e-9));
            if (max_clock < current_clock)
                max_clock = current_clock;

            VERBOSE (cpu, DEBUG, printf ("counts:"); for (i = 0; i < nb_cpu; i++) printf (" %d", counts[i]); printf ("\n"));
            VERBOSE (cpu, DEBUG, printf ("measurements:"); for (i = 0; i < nb_cpu; i++) printf (" %g", measurements[i]); printf ("\n"));

            /* build output string */
            char str_buffer[1024] = { 0 };
            int str_size = sizeof (str_buffer);
            char *str = str_buffer;
            str += snprintf (str, str_size, "cpu: ");
            str_size = sizeof (str_buffer) + (int)(str_buffer - str);

            /* for each cpu */
            float tot = 0.0f;
            for (i = 0; i < nb_cpu; i++) {

                /* if no measure done, output -0.0 */
                float c = (counts[i] <= 0) ? -1e-6f : 100.f * measurements[i] / (float)counts[i];

                /* estimate total load */
                tot += c;

                /* round */
                if ((int)(100 * c - (int)(10 * c) * 10) >= 5)
                    c += 0.05;

                /* add formated percentage */
                str += (c < 0) ? snprintf (str, str_size, " + ---.-%%") :
                    snprintf (str, str_size, "%s% 3d.%1d%%", i  ? " + " : "", (int)c, (int)(10 * (c - (int)c)));
                str_size = sizeof (str_buffer) + (int)(str_buffer - str);
            }

            /* add formated total percentage */
            str += snprintf (str, str_size, " = % 4d.%1d%%", (int)tot, (int)(10 * (tot - (int)tot)));

            /* print it */
            printf ("%s\n", str_buffer);

            /* check number of second since start */
            nb_seconds += timer;
            if ((0 < stop_after_nb_seconds) && (stop_after_nb_seconds <= nb_seconds)) {
                VERBOSE (cpu, WARNING, printf ("exiting after %d'' (threshold %d'')\n", nb_seconds, stop_after_nb_seconds));
                exit (0);
            }

            /* empty all tabs */
            memset (measurements, 0, nb_cpu * sizeof (float));
            memset (counts, 0, nb_cpu * sizeof (int));

            /* re-arm trigger */
            alarm_trigger = 0;
            alarm (timer);
            select_error = 0;
            ts1 = rdtsc ();
            gettimeofday (&tv1, NULL);
        }

        /* lets core sleep */
        usleep (1000);
    }
}

int main (int argc, char *argv[])
{
    int do_calibration = 0;
    int i;

    progname = basename (strdup (argv[0]));

    /* analyse parameters */
    int c;
    while ((c = getopt (argc, argv, "c:f:hk:m:s:t:v:V")) != EOF) {
        switch (c) {
        case 'c':
            do_calibration = 1;
            nb_calibration_loops = atoi (optarg);
            break;
        case 'f':
            calibration_file = strdup (optarg);
            break;
        case 'h':
        default:
            usage (c != 'h');
            break;
        case 'k':
            cpu_cal = atoi (optarg);
            break;
        case 'm':
            cpu_mon = atoi (optarg);
            break;
        case 's':
            stop_after_nb_seconds = atoi (optarg);
            break;
        case 't':
            timer = atoi (optarg);
            break;
        case 'v':
            CHANGE_VERBOSE_LEVEL(cpu, atoi (optarg));
            break;
        case 'V':
            printf ("%s version: %s\n", progname, VERSION);
            exit (0);
            break;
        }
    }
    if (argc - optind != 0) {
        fprintf (stderr, "%s: invalid option -- %s\n", progname, argv[optind]);
        usage (1);
    }

    /* get number of cpu */
    nb_cpu = get_cpu_number ();
    pipes = (int *) calloc (nb_cpu, sizeof (int));
    pids = (pid_t *) calloc (nb_cpu, sizeof (pid_t));
    VERBOSE (cpu, DEBUG, printf ("number of cpu: %d\n", nb_cpu));

    /* check cpu for calibration and monitoring */
    cpu_cal = cpu_cal < 0 ? nb_cpu + cpu_cal : cpu_cal;
    cpu_cal = (cpu_cal < 0) ? 0: (cpu_cal) >= nb_cpu ? nb_cpu - 1 : cpu_cal;
    cpu_mon = cpu_mon < 0 ? nb_cpu + cpu_mon : cpu_mon;
    cpu_mon = (cpu_mon < 0) ? 0: (cpu_mon) >= nb_cpu ? nb_cpu - 1 : cpu_mon;

    /* calibration */
    if (!do_calibration) {
        max_loops = read_calibration_file (calibration_file);
        max_clock = current_clock;

        if (max_loops <= 0)
            do_calibration = 1;
    }
    if ((do_calibration) && (nb_calibration_loops > 0)) {
        VERBOSE (cpu, WARNING, printf ("runnning calibration on cpu %d\n", cpu_cal));
        max_loops = calibrate (cpu_cal);
        max_clock = current_clock;

        write_calibration_file (calibration_file);
    }
    if (max_loops > 0)
        VERBOSE (cpu, WARNING, printf ("calibration max loops: %d\n", max_loops));

    /* sigpipe signal handler */
    struct sigaction act;
    memset (&act, 0, sizeof (act));
    act.sa_handler = sigpipe_handler;
    if (sigaction (SIGPIPE, &act, NULL))
        VERBOSE (cpu, WARNING, printf ("sigaction (SIGPIPE)\n"));

    /* fork for measurement */
    for (i = 0; i < nb_cpu; i++) {

        /* create pipe */
        int fds[2] = { 0 };
        if (pipe (fds)) {
            VERBOSE (cpu, ERROR, printf ("pipe: %s\n", strerror (errno)));
            exit (1);
        }

        /* fork process */
        pid_t pid = fork ();
        if (pid < 0) {
            VERBOSE (cpu, ERROR, printf ("fork: %s\n", strerror (errno)));
            exit (1);
        }

        /* child process */
        if (pid != 0) {
            int j;

            /* close useless fids */
            close (fds[0]);
            for (j = 0; j < i; j++)
                close (pipes[j]);
            pipes[i] = fds[1];

            /* measurement process */
            VERBOSE (cpu, DEBUG, printf ("running cpu measures on cpu%d\n", i));
            char buffer[32];
            snprintf (buffer, sizeof (buffer), "cpu-%02d", i);
            set_name (argc, argv, buffer);
            measurement_process (i);

            /* error in child process */
            VERBOSE (cpu, ERROR, printf ("error running cpu measures on cpu%d\n", i));
            exit (1);
        }

        /* close useless fid */
        close (fds[1]);
        pipes[i] = fds[0];

        /* keep chil pid */
        pids[i] = pid;
    }

    /* handler sigchild, sigint and kill all children at exit */
    act.sa_handler = sigchld_handler;
    if (sigaction (SIGCHLD, &act, NULL))
        VERBOSE (cpu, WARNING, printf ("sigaction (SIGCHLD)\n"));
    act.sa_handler = sigint_handler;
    if (sigaction (SIGINT, &act, NULL))
        VERBOSE (cpu, WARNING, printf ("sigaction (SIGINT)\n"));
    if (atexit (kill_all_children))
        VERBOSE (cpu, WARNING, printf ("cant set killall at exit\n"));

    /* monitoring process */
    VERBOSE (cpu, DEBUG, printf ("running monitoring on cpu%d\n", cpu_mon));
    set_name (argc, argv, "cpumon");
    monitoring_process (cpu_mon);

    /* error in father process */
    VERBOSE (cpu, ERROR, printf ("error running monitoring on cpu%d\n", i));
    return 0;
}

/* vi:set tabstop=4 expandtab shiftwidth=4: this line set vi mode */
