/*
  File name        : load.c
  Date of creation : 6/11/2013
  Version          : 1.2
  Copyright        : Soft'n'design
  Author           : Laurent Mazet <mazet@softndesign.org>

  Description      : This file contains cpu loading process

  History          :
  - initial version
*/

/* check OS */
#ifdef __FREEBSD__
#elif __LINUX__
#elif __NETBSD__
#elif __SYLLABLE__
#elif __DEPEND__
#else
#error "unsupported operating system"
#endif

#define _GNU_SOURCE

#include <assert.h>
#include <errno.h>
#include <getopt.h>
#include <libgen.h>
#include <pthread.h>
#ifdef __FREEBSD__
#include <pthread_np.h>
#endif
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __FREEBSD__
#include <sys/cpuset.h>
#endif
#ifdef __SYLLABLE__
typedef cpu_set_t cpu_set;
#endif
#include <sys/cdefs.h>
#include <sys/time.h>
#include <unistd.h>

#include "rdtsc.h"
#include "verbose.h"

#define VERSION "1.2"

/**
   Default values
*/
int cpu = 0;
double load = 10.0;
DECLARE_VERBOSE_LEVEL (load, 1);

/**
   Global variables*/
char *progname = NULL;

/**
   Display help message

   @param ret return code
*/
void usage (int ret)
{
    printf ("usage: %s [-c int] [-h] [-l flt] [-v int] [-V]\n", progname);
    printf (" -c int: cpu id (%d)\n", cpu);
    printf (" -h    : help message\n");
    printf (" -l flt: load percentage (%.1f%%), need root\n", load);
    printf (" -v int: verbose level\n");
    printf (" -V    : show version\n");

    exit (ret);
}

/**
   Handler for sigint

   @param signal
*/
void sigint_handler (int sig)
{
    if (sig == SIGINT) {
        VERBOSE (load, WARNING, printf ("ending load process\n"));
        exit (0);
    }
}

/**
   Alarm trigger
*/
int alarm_trigger = 0;

/**
   Handler for sigalarm

   @param signal
*/
void sigalrm_handler (int sig)
{
    if (sig == SIGALRM)
        alarm_trigger = 1;
}

int main (int argc, char *argv[])
{
    progname = basename (strdup (argv[0]));

    /* analyse parameter */
    int c;
    while ((c = getopt (argc, argv, "c:hl:v:V")) != EOF) {
        switch (c) {
        case 'c':
            cpu = atoi (optarg);
            break;
        case 'h':
            /* fall through */
        default:
            usage (c != 'h');
            break;
        case 'l':
            load = atof (optarg);
            break;
        case 'v':
            CHANGE_VERBOSE_LEVEL(load, atoi (optarg));
            break;
        case 'V':
            printf ("%s version: %s\n", progname, VERSION);
            exit (0);
            break;
        }
    }
    if (argc - optind != 0) {
        VERBOSE (load, ERROR, printf ("%s: invalid option -- %s\n", progname, argv[optind]));
        usage (1);
    }

    /* check load */
    if ((load < 1) || (load > 90)) {
        VERBOSE (load, ERROR, printf ("load out of bound [1, 90] (%g)\n", load));
        exit (1);
    }

    /* assign to a cpu */
#ifdef __NETBSD__
    cpuset_t *cpu_set;
    cpu_set = cpuset_create ();
    if (cpu_set == NULL) {
        VERBOSE (load, ERROR, printf ("cpuset_create: %s\n", strerror (errno)));
        exit (1);
    }
    cpuset_set (cpu, cpu_set);
    if (pthread_setaffinity_np (pthread_self (), cpuset_size (cpu_set), cpu_set) != 0) {
        VERBOSE (load, ERROR, printf ("pthread_setaffinity_np: %s\n", strerror (errno)));
        exit (1);
    }
#elif __FREEBSD__
    cpuset_t cpu_mask;
    CPU_ZERO (&cpu_mask);
    CPU_SET (cpu, &cpu_mask);
    if (pthread_setaffinity_np (pthread_self (), sizeof (cpuset_t), &cpu_mask) != 0) {
        VERBOSE (load, ERROR, printf ("pthread_setaffinity_np: %s\n", strerror (errno)));
        exit (1);
    }
#elif __SYLLABLE__
    cpu_set cpu_mask;
    __CPU_ZERO (&cpu_mask);
    __CPU_SET (cpu, &cpu_mask);
    if (sched_setaffinity (getpid (), sizeof (cpu_set), &cpu_mask) != 0) {
        /* Function not implemented */
        VERBOSE (load, ERROR, printf ("sched_setaffinity_np: %s\n", strerror (errno)));
        //exit (1);
    }
#elif __LINUX__
    cpu_set_t cpu_mask;
    CPU_ZERO (&cpu_mask);
    CPU_SET (cpu, &cpu_mask);
    if (pthread_setaffinity_np (pthread_self (), sizeof (cpu_set_t), &cpu_mask) != 0) {
        VERBOSE (load, ERROR, printf ("pthread_setaffinity_np: %s\n", strerror (errno)));
        exit (1);
    }
#endif

    /* activate realtime scheduler */
    struct sched_param sp;
    memset (&sp, 0, sizeof (sp));
    sp.sched_priority = sched_get_priority_max (SCHED_FIFO) - 10;
    /* max prio can freeze linux kernel, minus 10 is fine */
    if (sched_setscheduler (0, SCHED_FIFO, &sp) < 0) {
        VERBOSE (load, ERROR, printf ("sched_setscheduler: %s\n", strerror (errno)));
        exit (1);
    }

    /* signal handler for sigint */
    struct sigaction act;
    memset (&act, 0, sizeof (act));
    act.sa_handler = sigint_handler;
    if (sigaction (SIGINT, &act, NULL))
        VERBOSE (load, WARNING, printf ("sigaction (SIGINT)\n"));

    /* signal handler for sigalarm */
    memset (&act, 0, sizeof (act));
    act.sa_handler = sigalrm_handler;
    if (sigaction (SIGALRM, &act, NULL))
        VERBOSE (load, WARNING, printf ("sigaction (SIGALRM)\n"));

    /* set alarm */
    alarm_trigger = 0;
    alarm (1);

    /* calibration */
    VERBOSE (load, DEBUG, printf ("starting calibration\n"));

    /* get time */
    struct timeval tv1;
    gettimeofday (&tv1, NULL);
    int64_t ts1 = rdtsc ();

    /* loop counter */
    int cpt = -1;
    int64_t min_ts = INT64_MAX;
    while (!alarm_trigger) {
        int k, sum = 0;
        int64_t ts = rdtsc ();
        for (k = 0; k < 10000; k++) sum += k;
        ts = rdtsc () - ts;
        if (min_ts > ts)
            min_ts = ts;
        cpt++;
    }

    /* get time */
    int64_t ts2 = rdtsc ();
    struct timeval tv2;
    gettimeofday (&tv2, NULL);

    /* cpu clock estimate */
    double delta = ((double)tv2.tv_sec + tv2.tv_usec * 1e-6) -
        ((double)tv1.tv_sec + tv1.tv_usec * 1e-6);
    double cpu_clock = (ts2 - ts1) / delta;
    int cal = cpt;
    VERBOSE (load, DEBUG, printf ("cpu%d: %.3fGHz %d, loop in %.0f ms\n",
                                  cpu, cpu_clock * 1e-9, cal, delta * 1e3));

    /* main loop */
    while (1) {

        /* load loop */
        int i;
        ts1 = rdtsc ();
        for (i = 0; i < cal / 100; i ++) {
            int k, sum = 0;
            for (k = 0;k < 10000; k++) sum += k;
        }
        ts2 = rdtsc ();

        /* re-estimate */
        delta = (ts2 - ts1) / cpu_clock * 1e6;

        /* sleep */
        usleep (delta * (100 - load) / load);

        VERBOSE (load, DEBUG, printf ("delta: %g %g\n", delta, delta * (100 - load) / load));
    }

    return 0;
}

/* vi:set tabstop=4 expandtab shiftwidth=4: this line set vi mode */
